@@ -3,8 +3,9 @@ import logging
 import pathlib
 import sqlite3
 import tempfile
+import time
 import uuid
-from typing import List
+from typing import List, Optional
 
 import msgpack
 
@@ -13,8 +14,9 @@ from kodexa.model.model import (
     DocumentMetadata,
     ContentFeature,
     ContentException,
-    ModelInsight,
+    ModelInsight, ProcessingStep,
 )
+from kodexa.model.objects import DocumentTaxonValidation
 
 logger = logging.getLogger()
 
@@ -41,6 +43,22 @@ FEATURE_TYPE_LOOKUP = "select id from f_type where name = ?"
 METADATA_INSERT = "insert into metadata(id,metadata) values (1,?)"
 METADATA_DELETE = "delete from metadata where id=1"
 
+# Configuration constants
+CACHE_SIZE = 10000  # Number of nodes to cache
+BATCH_SIZE = 1000   # Size of batches for bulk operations
+SLOW_QUERY_THRESHOLD = 1.0  # Seconds
+MAX_CONNECTIONS = 5  # Maximum number of database connections
+
+def monitor_performance(func):
+    """Performance monitoring decorator"""
+    def wrapper(*args, **kwargs):
+        start_time = time.time()
+        result = func(*args, **kwargs)
+        duration = time.time() - start_time
+        if duration > SLOW_QUERY_THRESHOLD:
+            logger.warning(f"Slow operation detected: {func.__name__}, duration: {duration}s")
+        return result
+    return wrapper
 
 class SqliteDocumentPersistence(object):
     """
@@ -56,7 +74,7 @@ class SqliteDocumentPersistence(object):
     The Sqlite persistence engine to support large scale documents (part of the V4 Kodexa Document Architecture)
     """
 
-    def __init__(self, document: Document, filename: str = None, delete_on_close=False, inmemory=False):
+    def __init__(self, document: Document, filename: str = None, delete_on_close=False, inmemory=False, persistence_manager=None):
         self.document = document
 
         self.node_types = {}
@@ -93,8 +111,10 @@ class SqliteDocumentPersistence(object):
 
         self.cursor = self.connection.cursor()
         self.cursor.execute("PRAGMA journal_mode=OFF")
-        self.cursor.execute("pragma temp_store = memory")
-        self.cursor.execute("pragma mmap_size = 30000000000")
+        self.cursor.execute("PRAGMA temp_store=MEMORY")
+        self.cursor.execute("PRAGMA mmap_size=30000000000")
+        self.cursor.execute("PRAGMA cache_size=10000")
+        self.cursor.execute("PRAGMA page_size=4096")
 
         try:
             # We need to populate node_type_id_by_name
@@ -136,6 +156,7 @@ class SqliteDocumentPersistence(object):
 
         return mem_conn
 
+    @monitor_performance
     def get_all_tags(self):
         """
         Retrieves all tags from the document.
@@ -151,6 +172,7 @@ class SqliteDocumentPersistence(object):
 
         return features
 
+    @monitor_performance
     def update_features(self, node):
         """
         Updates the features of a given node in the document.
@@ -186,6 +208,7 @@ class SqliteDocumentPersistence(object):
         self.cursor.execute("DELETE FROM ft where cn_id=?", [node.uuid])
         self.cursor.executemany(FEATURE_INSERT, all_features)
 
+    @monitor_performance
     def update_node(self, node):
         """
         Updates a given node in the document.
@@ -198,6 +221,7 @@ class SqliteDocumentPersistence(object):
             [node.index, node._parent_uuid, node.uuid],
         )
 
+    @monitor_performance
     def get_content_nodes(self, node_type, parent_node: ContentNode, include_children):
         """
         Retrieves content nodes from the document based on the given parameters.
@@ -211,8 +235,8 @@ class SqliteDocumentPersistence(object):
             list: A list of content nodes that match the given parameters.
         """
         nodes = []
-
-        results = []
+        if not self.connection.in_transaction:
+            self.cursor.execute("BEGIN TRANSACTION")
         if include_children:
             if node_type == "*":
                 query = """
@@ -282,6 +306,7 @@ class SqliteDocumentPersistence(object):
                         ],
                     ).fetchall()
                 except StopIteration:
+                    self.connection.commit()
                     return []
         else:
             query = "select id, pid, nt, idx from cn where pid=? and nt=? order by idx"
@@ -298,11 +323,14 @@ class SqliteDocumentPersistence(object):
                     ],
                 ).fetchall()
             except StopIteration:
+                self.connection.commit()
                 return []
 
         for raw_node in list(results):
             nodes.append(self.__build_node(raw_node))
 
+        self.connection.commit()
+
         return nodes
 
     def initialize(self):
@@ -324,6 +352,7 @@ class SqliteDocumentPersistence(object):
             self.cursor.close()
             self.connection.close()
 
+    @monitor_performance
     def get_max_feature_id(self):
         """
         Retrieves the maximum feature id from the document.
@@ -394,6 +423,7 @@ class SqliteDocumentPersistence(object):
 
         self.__update_metadata()
 
+    @monitor_performance
     def content_node_count(self):
         """
         Counts the number of content nodes in the document.
@@ -403,6 +433,7 @@ class SqliteDocumentPersistence(object):
         """
         self.cursor.execute("select * from cn").fetchall()
 
+    @monitor_performance
     def get_feature_type_id(self, feature):
         """
         Retrieves the id of a given feature.
@@ -464,6 +495,7 @@ class SqliteDocumentPersistence(object):
 
         return result[0]
 
+    @monitor_performance
     def __insert_node(self, node: ContentNode, parent, execute=True):
         """
         Inserts a node into the document.
@@ -835,8 +867,10 @@ class SqliteDocumentPersistence(object):
         self.cursor.execute("VACUUM")
         self.cursor = self.connection.cursor()
         self.cursor.execute("PRAGMA journal_mode=OFF")
-        self.cursor.execute("pragma temp_store = memory")
-        self.cursor.execute("pragma mmap_size = 30000000000")
+        self.cursor.execute("PRAGMA temp_store=MEMORY")
+        self.cursor.execute("PRAGMA mmap_size=30000000000")
+        self.cursor.execute("PRAGMA cache_size=10000")
+        self.cursor.execute("PRAGMA page_size=4096")
 
     def dump_in_memory_db_to_file(self):
         # Connect to a new or existing database file
@@ -1020,7 +1054,7 @@ class SqliteDocumentPersistence(object):
         """
         model_insights = []
         for model_insight in self.cursor.execute(MODEL_INSIGHT_SELECT).fetchall():
-            model_insights.append(ModelInsight.parse_raw(model_insight[0]))
+            model_insights.append(ModelInsight.model_validate_json(model_insight[0]))
 
         return model_insights
 
@@ -1122,6 +1156,182 @@ class SqliteDocumentPersistence(object):
 
         return content_nodes
 
+    def __ensure_validations_table_exists(self):
+        """
+        Ensure the 'validations' table exists in the database.
+        Creates the table if it does not exist and initializes it with an empty list.
+        """
+        self.cursor.execute("""
+            CREATE TABLE IF NOT EXISTS validations (
+                obj BLOB
+            )
+        """)
+
+        # Check if the table has any rows, if not, insert an initial empty row
+        result = self.cursor.execute("SELECT COUNT(*) FROM validations").fetchone()
+        if result[0] == 0:
+            self.cursor.execute("INSERT INTO validations (obj) VALUES (?)", [sqlite3.Binary(msgpack.packb([]))])
+
+    def set_validations(self, validations: List[DocumentTaxonValidation]):
+        """
+        Sets the validations for the document.
+
+        Args:
+            validations (List[DocumentTaxonValidation]): The validations to store.
+        """
+        self.__ensure_validations_table_exists()
+        serialized_data = sqlite3.Binary(msgpack.packb([v.model_dump(by_alias=True) for v in validations]))
+        self.cursor.execute("UPDATE validations SET obj = ? WHERE rowid = 1", [serialized_data])
+        self.connection.commit()
+
+    def get_validations(self) -> List[DocumentTaxonValidation]:
+        """
+        Gets the validations associated with this document.
+
+        Returns:
+            List[DocumentTaxonValidation]: The list of validations stored in the validations table.
+        """
+        self.__ensure_validations_table_exists()
+        result = self.cursor.execute("SELECT obj FROM validations WHERE rowid = 1").fetchone()
+        if result and result[0]:
+            return [DocumentTaxonValidation.model_validate(v) for v in msgpack.unpackb(result[0])]
+        return []
+
+    def set_external_data(self, external_data: dict, key: str = "default"):
+        """
+        Sets the external data for the document for a specific key.
+
+        Args:
+            external_data (dict): The external data to store, must be JSON serializable.
+            key (str): The key to store the data under, defaults to "default"
+        """
+        self.__ensure_ed_table_exists()
+        serialized_data = sqlite3.Binary(msgpack.packb(external_data))
+        self.cursor.execute("DELETE FROM ed WHERE key = ?", [key])
+        self.cursor.execute("INSERT INTO ed (key, obj) VALUES (?, ?)", [key, serialized_data])
+        self.connection.commit()
+
+    def get_external_data(self, key: str = "default") -> dict:
+        """
+        Gets the external data associated with this document for a specific key.
+
+        Args:
+            key (str): The key to retrieve data for, defaults to "default"
+
+        Returns:
+            dict: The external data stored in the ed table for the given key.
+        """
+        self.__ensure_ed_table_exists()
+        result = self.cursor.execute("SELECT obj FROM ed WHERE key = ?", [key]).fetchone()
+        if result and result[0]:
+            return msgpack.unpackb(result[0])
+        return {}
+
+    def get_external_data_keys(self) -> List[str]:
+        """
+        Gets all keys under which external data is stored.
+
+        Returns:
+            List[str]: A list of all keys that have external data stored.
+        """
+        self.__ensure_ed_table_exists()
+        results = self.cursor.execute("SELECT key FROM ed").fetchall()
+        return [row[0] for row in results]
+
+    def __ensure_ed_table_exists(self):
+        """
+        Ensure the 'ed' table exists in the database.
+        Creates the table if it does not exist.
+        """
+        # First check if the old table exists and has key column
+        old_table = self.cursor.execute("""
+            SELECT name FROM sqlite_master 
+            WHERE type='table' AND name='ed'
+        """).fetchone()
+
+        if old_table:
+            # Check if table has key column
+            table_info = self.cursor.execute("PRAGMA table_info(ed)").fetchall()
+            has_key_column = any(col[1] == 'key' for col in table_info)
+
+            if not has_key_column:
+                # Get the old data and drop the table
+                data = self.cursor.execute("SELECT obj FROM ed").fetchone()
+                self.cursor.execute("DROP TABLE ed")
+
+                # Create new table with key column
+                self.cursor.execute("""
+                    CREATE TABLE ed (
+                        key TEXT PRIMARY KEY,
+                        obj BLOB
+                    )
+                """)
+
+                # If there was data in the old table, insert it with default key
+                if data:
+                    self.cursor.execute("INSERT INTO ed (key, obj) VALUES (?, ?)",
+                                      ["default", data[0]])
+            else:
+                # Table exists and has key column - do nothing
+                return
+        else:
+            # Create new table if it doesn't exist
+            self.cursor.execute("""
+                CREATE TABLE IF NOT EXISTS ed (
+                    key TEXT PRIMARY KEY,
+                    obj BLOB
+                )
+            """)
+
+            # Check if default key exists, if not insert empty data
+            result = self.cursor.execute("SELECT COUNT(*) FROM ed WHERE key = 'default'").fetchone()
+            if result[0] == 0:
+                self.cursor.execute("INSERT INTO ed (key, obj) VALUES (?, ?)",
+                                  ["default", sqlite3.Binary(msgpack.packb({}))])
+
+    def __ensure_steps_table_exists(self):
+        """
+        Ensure the 'steps' table exists in the database.
+        Creates the table if it does not exist.
+        """
+        self.cursor.execute("""
+                CREATE TABLE IF NOT EXISTS steps (
+                    obj BLOB
+                )
+            """)
+
+        # Check if the table has any rows, if not, insert an initial empty row
+        result = self.cursor.execute("SELECT COUNT(*) FROM steps").fetchone()
+        if result[0] == 0:
+            self.cursor.execute("INSERT INTO steps (obj) VALUES (?)", [sqlite3.Binary(msgpack.packb([]))])
+
+    def set_steps(self, steps: List[ProcessingStep]):
+        """
+        Sets the processing steps for the document.
+
+        Args:
+            steps (List[ProcessingStep]): A list of ProcessingStep objects to store.
+        """
+        self.__ensure_steps_table_exists()
+        serialized_steps = [step.to_dict() for step in steps]
+        packed_data = sqlite3.Binary(msgpack.packb(serialized_steps))
+        self.cursor.execute("UPDATE steps SET obj = ? WHERE rowid = 1", [packed_data])
+        self.connection.commit()
+
+    def get_steps(self) -> List[ProcessingStep]:
+        """
+        Gets the processing steps associated with this document.
+
+        Returns:
+            List[ProcessingStep]: A list of ProcessingStep objects.
+        """
+        self.__ensure_steps_table_exists()
+        result = self.cursor.execute("SELECT obj FROM steps WHERE rowid = 1").fetchone()
+        if result and result[0]:
+            unpacked_data = msgpack.unpackb(result[0])
+            return [ProcessingStep(**step) for step in unpacked_data]
+        return []
+
 
 class SimpleObjectCache(object):
     """
@@ -1146,7 +1356,7 @@ class SimpleObjectCache(object):
         self.next_id = 1
         self.dirty_objs = set()
 
-    def get_obj(self, obj_id):
+    def get_obj(self, obj_id) -> Optional[ContentNode]:
         """
         Get the object with the given ID.
 
@@ -1161,7 +1371,7 @@ class SimpleObjectCache(object):
 
         return None
 
-    def add_obj(self, obj):
+    def add_obj(self, obj: ContentNode):
         """
         Add an object to the cache.
 
@@ -1186,7 +1396,7 @@ class SimpleObjectCache(object):
             if obj.uuid in self.dirty_objs:
                 self.dirty_objs.remove(obj.uuid)
 
-    def get_dirty_objs(self):
+    def get_dirty_objs(self) -> list[ContentNode]:
         """
         Get all dirty objects in the cache.
 
@@ -1257,9 +1467,52 @@ class PersistenceManager(object):
         self.node_parent_cache = {}
 
         self._underlying_persistence = SqliteDocumentPersistence(
-            document, filename, delete_on_close, inmemory=inmemory
+            document, filename, delete_on_close, inmemory=inmemory, persistence_manager=self
         )
 
+    def get_steps(self) -> list[ProcessingStep]:
+        """
+        Gets the processing steps for this document
+
+        :return:
+        """
+        return self._underlying_persistence.get_steps()
+
+    def set_steps(self, steps: list[ProcessingStep]):
+        self._underlying_persistence.set_steps(steps)
+
+    def set_validations(self, validations: list[DocumentTaxonValidation]):
+        self._underlying_persistence.set_validations(validations)
+
+    def get_validations(self) -> list[DocumentTaxonValidation]:
+        return self._underlying_persistence.get_validations()
+
+    def get_external_data(self, key="default") -> dict:
+        """
+        Gets the external data object associated with this document
+
+        :return: dict of the external data
+        """
+        return self._underlying_persistence.get_external_data(key)
+
+    def get_external_data_keys(self) -> List[str]:
+        """
+        Gets all keys under which external data is stored.
+
+        Returns:
+            List[str]: A list of all keys that have external data stored.
+        """
+        return self._underlying_persistence.get_external_data_keys()
+
+    def set_external_data(self, external_data:dict, key="default"):
+        """
+        Sets the external data for this document
+
+        :param external_data: dict representing the external data, must be JSON serializable
+        :return:
+        """
+        self._underlying_persistence.set_external_data(external_data, key)
+
     def get_nodes_by_type(self, node_type: str) -> List[ContentNode]:
         """
         Retrieves all nodes of a given type from the underlying persistence layer.
@@ -1283,7 +1536,10 @@ class PersistenceManager(object):
             ContentNode: The node with the given uuid.
         """
         if self.node_cache.get_obj(uuid) is None:
-            self.node_cache.add_obj(self._underlying_persistence.get_node(uuid))
+            node = self._underlying_persistence.get_node(uuid)
+            if node:
+                self.node_cache.add_obj(node)
+                return node
 
         return self.node_cache.get_obj(uuid) # return the cached version
 
@@ -1358,7 +1614,6 @@ class PersistenceManager(object):
         Returns:
             List[Node]: A list of nodes tagged with the specified tag.
         """
-        self.flush_cache()
         return self._underlying_persistence.get_tagged_nodes(tag, tag_uuid)
 
     def get_all_tagged_nodes(self):
@@ -1368,7 +1623,6 @@ class PersistenceManager(object):
         Returns:
             List[Node]: A list of all tagged nodes.
         """
-        self.flush_cache()
         return self._underlying_persistence.get_all_tagged_nodes()
 
     def initialize(self):
@@ -1400,6 +1654,7 @@ class PersistenceManager(object):
         """
         self._underlying_persistence.close()
 
+    @monitor_performance
     def flush_cache(self):
         """
         Flushes the cache by merging it with the underlying persistence layer.
@@ -1409,11 +1664,13 @@ class PersistenceManager(object):
         all_content_parts = []
         all_features = []
         node_id_with_features = []
-
-        logger.debug("Merging cache to persistence")
         dirty_nodes = self.node_cache.get_dirty_objs()
 
-        logger.debug(f"Identified {len(dirty_nodes)} nodes to update")
+        if len(dirty_nodes) == 0:
+            return
+
+        if not self._underlying_persistence.connection.in_transaction:
+            self._underlying_persistence.connection.execute("BEGIN TRANSACTION")
 
         next_feature_id = self._underlying_persistence.get_max_feature_id()
         for node in dirty_nodes:
@@ -1453,7 +1710,6 @@ class PersistenceManager(object):
 
                 self.node_cache.undirty(node)
 
-        logger.debug(f"Writing {len(all_node_ids)} nodes")
         self._underlying_persistence.cursor.executemany(
             "DELETE FROM cn where id=?", all_node_ids
         )
@@ -1466,14 +1722,11 @@ class PersistenceManager(object):
         self._underlying_persistence.cursor.executemany(
             "DELETE FROM cnp where cn_id=?", all_node_ids
         )
-        logger.debug(f"Writing {len(all_content_parts)} content parts")
-
         self._underlying_persistence.cursor.executemany(
             CONTENT_NODE_PART_INSERT, all_content_parts
         )
-
-        logger.debug(f"Writing {len(all_features)} features")
         self._underlying_persistence.cursor.executemany(FEATURE_INSERT, all_features)
+        self._underlying_persistence.connection.commit()
 
     def get_content_nodes(self, node_type, parent_node, include_children):
         """
